from typing import Iterable, Dict, List
from uuid import UUID

from openpyxl import load_workbook

from sourcesquirrel.classes.collection import Collection
from sourcesquirrel.classes.media_type import MediaType
from sourcesquirrel.classes.release_type import ReleaseType
from sourcesquirrel.classes.serie import Serie
from sourcesquirrel.constants.blockchains import ALGORAND, CARDANO, COINBASE, ETHEREUM, POLYGON
from sourcesquirrel.prefabs import blockchains, platforms
from sourcesquirrel.validators import algorand, cardano, evm

COLLECTIONS_SHEET = "collections"
SERIES_SHEET = "series"
MEDIA_TYPES_SHEET = "media types"
RELEASE_TYPES_SHEET = "release types"

ONCHAIN_COLLECTION_ID_VALIDATORS = {
    ALGORAND: algorand.is_address,
    CARDANO: cardano.is_policy_id,
    COINBASE: evm.is_address,
    ETHEREUM: evm.is_address,
    POLYGON: evm.is_address,
}


def read_rows(file_path: str, sheet_name: str, skip_header: bool = True):
    workbook = load_workbook(file_path, data_only=True)

    if sheet_name not in workbook.sheetnames:
        raise ValueError(f"Sheet '{sheet_name}' does not exist in the workbook.")

    sheet = workbook[sheet_name]
    rows = sheet.iter_rows(values_only=True)

    if skip_header:
        next(rows)

    for row in rows:
        if row[0] is None:
            break

        yield row


def parse_list(value: str) -> List[str]:
    return list(filter(bool, map(str.strip, (value or "").split(","))))


def load_series(path: str) -> Iterable[Serie]:
    for row in read_rows(path, SERIES_SHEET):
        id = row[0]
        slug = row[1]
        name = row[2]
        emoji = row[3]

        blockchain = blockchains.get(row[4])

        serie = Serie(
            id=UUID(id),
            name=name,
            slug=slug,
            emoji=emoji,
            blockchain=blockchain,
        )

        serie.verify()
        yield serie


def load_media_types(path: str) -> Iterable[MediaType]:
    for row in read_rows(path, MEDIA_TYPES_SHEET):
        id = row[0]
        name = row[1]
        emoji = row[2]
        platform = platforms.get(row[3])

        media_type = MediaType(
            id=UUID(id),
            name=name,
            emoji=emoji,
            platform=platform,
        )

        media_type.verify()
        yield media_type


def load_release_types(path: str) -> Iterable[ReleaseType]:
    for row in read_rows(path, RELEASE_TYPES_SHEET):
        id = row[0]
        name = row[1]
        emoji = row[2]

        release_type = ReleaseType(
            id=UUID(id),
            name=name,
            emoji=emoji,
        )

        release_type.verify()
        yield release_type


def load_collections(
    path: str, all_series: Dict[str, Serie], all_media_types: Dict[str, MediaType], all_release_types: Dict[str, ReleaseType]
) -> Iterable[Collection]:
    for row in read_rows(path, COLLECTIONS_SHEET):
        id = row[0]
        blockchain = blockchains.get(row[1])
        media_type = all_media_types.get(row[2])
        onchain_collection_id = row[3]

        if "/" in onchain_collection_id:
            onchain_collection_id, onchain_collection_id_mainnet = onchain_collection_id.split("/")
        else:
            onchain_collection_id_mainnet = onchain_collection_id

        onchain_discriminator = row[4]
        title = row[5]
        is_public_domain = row[6]
        release_date = row[7].date()
        supply = int(row[8])
        cover_count = int(row[9])
        one_to_one_count = int(row[10])
        team_royalties = float(row[11])
        source_royalties = float(row[12])
        release_type = all_release_types.get(row[13])

        authors = parse_list(row[14])
        publishers = parse_list(row[15])
        series = [all_series[slug] for slug in parse_list(row[16])]

        collection = Collection(
            id=UUID(id),
            blockchain=blockchain,
            media_type=media_type,
            onchain_collection_id=onchain_collection_id,
            onchain_collection_id_mainnet=onchain_collection_id_mainnet,
            onchain_discriminator=onchain_discriminator,
            title=title,
            is_public_domain=is_public_domain,
            release_date=release_date,
            supply=supply,
            cover_count=cover_count,
            one_to_one_count=one_to_one_count,
            team_royalties=team_royalties,
            source_royalties=source_royalties,
            release_type=release_type,
            authors=authors,
            publishers=publishers,
            series=series,
        )

        collection.verify()
        yield collection



import logging
from typing import Iterable, List, Any

from sourcesquirrel import excel
from sourcesquirrel.classes.blockchain import Blockchain
from sourcesquirrel.classes.collection import Collection
from sourcesquirrel.classes.media_type import MediaType
from sourcesquirrel.classes.platform import Platform
from sourcesquirrel.classes.release_type import ReleaseType
from sourcesquirrel.classes.serie import Serie
from sourcesquirrel.integrations.google.drive import GoogleDriveClient
from sourcesquirrel.prefabs import blockchains, platforms

logger = logging.getLogger(__name__)


class SourceOfTruth:
    def __init__(
        self,
        collections: List[Collection],
        series: List[Serie],
        media_types: List[MediaType],
        release_types: List[ReleaseType],
    ):
        self.collections: List[Collection] = collections
        self.series: List[Serie] = series
        self.media_types: List[MediaType] = media_types
        self.release_types: List[ReleaseType] = release_types

    @property
    def blockchains(self) -> Iterable[Blockchain]:
        return list(blockchains.BLOCKCHAINS.values())

    @property
    def platforms(self) -> Iterable[Platform]:
        return list(platforms.PLATFORMS.values())

    def verify(self):
        logger.info("blockchains")

        self.verify_helper("blockchain", self.blockchains)
        self.verify_helper("collection", self.collections)
        self.verify_helper("media_type", self.media_types)
        self.verify_helper("platform", self.platforms)
        self.verify_helper("release_type", self.release_types)
        self.verify_helper("serie", self.series)

    def verify_helper(self, name: str, items: List[Any]):
        logger.info(f"{name}s")

        ids = set()

        for item in items:
            if item.id in ids:
                raise ValueError(f"Duplicate {name} id: {item.id}")

            try:
                item.verify()
                ids.add(item.id)

                logger.info("* %s", item)
            except Exception as e:
                raise ValueError(f"Error verifying {name} {item.id}: {e}")

        pass

    @staticmethod
    def load_xlsx(path: str) -> "SourceOfTruth":
        series = {s.slug: s for s in excel.load_series(path)}
        media_types = {mt.name: mt for mt in excel.load_media_types(path)}
        release_types = {rt.name: rt for rt in excel.load_release_types(path)}
        collections = [c for c in excel.load_collections(path, series, media_types, release_types)]

        return SourceOfTruth(
            collections=collections,
            series=list(series.values()),
            media_types=list(media_types.values()),
            release_types=list(release_types.values()),
        )

    @staticmethod
    def load_from_drive(
        file_drive_id: str,
        file_path: str = "/tmp/sot.xlsx",
        google_secrets_json: str = None,
        google_secrets_path: str = None,
    ) -> "SourceOfTruth":
        if google_secrets_json:
            client = GoogleDriveClient.from_google_secrets_json(google_secrets_json)
        elif google_secrets_path:
            client = GoogleDriveClient.from_google_secrets_path(google_secrets_path)
        else:
            raise ValueError("Missing Google secrets")

        client.download(file_drive_id, file_path)

        return SourceOfTruth.load_xlsx(file_path)



from sourcesquirrel.constants.blockchains import ALGORAND, CARDANO, COINBASE, ETHEREUM, POLYGON
from sourcesquirrel.helpers import required
from sourcesquirrel.validators import algorand, cardano, evm


class Blockchain:
    def __init__(self, name: str, currency: str):
        self.id: str = name
        self.name: str = name
        self.currency: str = currency

    def is_collection_id(self, collection_id: str) -> bool:
        if self.name == ALGORAND:
            return algorand.is_address(collection_id)
        elif self.name == CARDANO:
            return cardano.is_policy_id(collection_id)
        elif self.name in (COINBASE, ETHEREUM, POLYGON):
            return evm.is_address(collection_id)
        else:
            raise ValueError(f"Unknown blockchain: {self.name}")

    def verify(self):
        required("blockchain.name", self.name, str)
        required("blockchain.currency", self.currency, str)

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.name} ({self.currency})"



from sourcesquirrel.classes.blockchain import Blockchain
from sourcesquirrel.constants import blockchains as b
from sourcesquirrel.constants import cryptocurrencies as c

ALGORAND = Blockchain(b.ALGORAND, c.ALGO)
CARDANO = Blockchain(b.CARDANO, c.ADA)
COINBASE = Blockchain(b.COINBASE, c.BASE)
ETHEREUM = Blockchain(b.ETHEREUM, c.ETH)
POLYGON = Blockchain(b.POLYGON, c.POL)

BLOCKCHAINS = {
    b.ALGORAND: ALGORAND,
    b.CARDANO: CARDANO,
    b.COINBASE: COINBASE,
    b.ETHEREUM: ETHEREUM,
    b.POLYGON: POLYGON,
}


def get(name: str) -> Blockchain:
    return BLOCKCHAINS[name]



from typing import Tuple, Optional


def split_asset_id(asset_id: str) -> Tuple[str, str, Optional[str]]:
    try:
        return asset_id[:56], asset_id[56:], bytes.fromhex(asset_id[56:]).decode()
    except UnicodeDecodeError:
        return asset_id[:56], asset_id[56:], None



from uuid import UUID

from sourcesquirrel.helpers import required


class ReleaseType:
    def __init__(self, id: UUID, name: str, emoji: str):
        self.id: UUID = id
        self.name: str = name
        self.emoji: str = emoji

    def verify(self):
        required("release_type.name", self.name, str)
        required("release_type.emoji", self.emoji, str)

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.emoji} {self.name}"



from sourcesquirrel.validators.utils import check


ADDRESS_REGEX = r"^0x[a-f0-9]{40}$"
ASSET_ID_REGEX = r"^[0-9]+$"
ENS_DOMAIN_REGEX = r"^[a-z0-9]+([\-\.][a-z0-9]+)*\.eth$"


def is_address(string: str) -> bool:
    return check(ADDRESS_REGEX, string)


def is_asset_id(string: str) -> bool:
    return check(ASSET_ID_REGEX, string)


def is_ens_domain(string: str) -> bool:
    return check(ENS_DOMAIN_REGEX, string)


def is_account(wallet: str) -> bool:
    return is_address(wallet) or is_ens_domain(wallet)



from typing import Any


def check(name: str, value: Any, type_: type, required_: bool):
    if required_:
        assert value is not None, f"{name} is required"

        if isinstance(value, str):
            assert value != "", f"{name} is empty"
    elif value is None:
        return

    assert isinstance(value, type_), f"{name} must be a {type_.__name__} and not {type(value).__name__}"


def required(name: str, value: Any, type_: type):
    check(name, value, type_, required_=True)


def optional(name: str, value: Any, type_: type):
    check(name, value, type_, required_=False)



from uuid import UUID

from sourcesquirrel.classes.blockchain import Blockchain
from sourcesquirrel.helpers import required


class Serie:
    def __init__(self, id: UUID, name: str, slug: str, emoji: str, blockchain: Blockchain):
        self.id: UUID = id
        self.name: str = name
        self.slug: str = slug
        self.emoji: str = emoji
        self.blockchain: Blockchain = blockchain

    def verify(self):
        required("serie.id", self.id, UUID)
        required("serie.name", self.name, str)
        required("serie.slug", self.slug, str)
        required("serie.emoji", self.emoji, str)
        required("serie.blockchain", self.blockchain, Blockchain)

        self.blockchain.verify()

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.emoji} {self.name} ({self.blockchain.name})"



from sourcesquirrel.constants.blockchains import ALGORAND, CARDANO, COINBASE, ETHEREUM, POLYGON
from . import algorand, cardano, evm


def is_collection_id(blockchain_name: str, value: str) -> bool:
    if blockchain_name == ALGORAND:
        return algorand.is_address(value)
    if blockchain_name == CARDANO:
        return cardano.is_policy_id(value)
    if blockchain_name in (COINBASE, ETHEREUM, POLYGON):
        return evm.is_address(value)

    raise ValueError(f"Unknown blockchain: {blockchain_name}")



from typing import Optional
from uuid import UUID

from sourcesquirrel.classes.platform import Platform
from sourcesquirrel.helpers import required, optional


class MediaType:
    def __init__(self, id: UUID, name: str, emoji: str, platform: Platform = None):
        self.id: UUID = id
        self.name: str = name
        self.emoji: str = emoji
        self.platform: Optional[Platform] = platform

    def verify(self) -> bool:
        required("media_type.id", self.id, UUID)
        required("media_type.name", self.name, str)
        required("media_type.emoji", self.emoji, str)
        optional("media_type.platform", self.platform, Platform)

        if self.platform:
            self.platform.verify()

    def __str__(self) -> str:  # pragma: no cover
        if self.platform is None:
            return f"{self.emoji} {self.name}"

        return f"{self.emoji} {self.name} ({self.platform.name})"



from sourcesquirrel.classes.platform import Platform
from sourcesquirrel.constants import platforms as p

BOOK_IO = Platform(p.BOOK_IO)
STUFF_IO = Platform(p.STUFF_IO)

PLATFORMS = {
    p.BOOK_IO: BOOK_IO,
    p.STUFF_IO: STUFF_IO,
}


def get(name: str) -> Platform:
    return PLATFORMS.get(name)



import logging

from oauth2client.service_account import ServiceAccountCredentials
from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive

SCOPES = ["https://www.googleapis.com/auth/drive"]
DEFAULT_SECRETS_PATH = "/etc/google/secrets.json"

logger = logging.getLogger(__name__)


class GoogleDriveClient:
    def __init__(self, secrets_path: str):
        secrets_path = secrets_path or DEFAULT_SECRETS_PATH

        gauth = GoogleAuth()
        gauth.credentials = ServiceAccountCredentials.from_json_keyfile_name(
            filename=secrets_path,
            scopes=SCOPES,
        )

        self.drive: GoogleDrive = GoogleDrive(gauth)

    @staticmethod
    def from_google_secrets_path(google_secrets_path: str) -> "GoogleDriveClient":
        return GoogleDriveClient(
            secrets_path=google_secrets_path,
        )

    @staticmethod
    def from_google_secrets_json(google_secrets_json: str) -> "GoogleDriveClient":
        google_secrets_path = "/tmp/secrets.json"

        with open(google_secrets_path, "w") as f:
            f.write(google_secrets_json)

        return GoogleDriveClient.from_google_secrets_path(
            google_secrets_path=google_secrets_path,
        )

    def download(self, file_id: str, path: str) -> None:
        file = self.drive.CreateFile({"id": file_id})
        file.GetContentFile(path)

        logger.info(f"Downloaded file {file_id} to {path}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sourcesquirrel',
 'sourcesquirrel.classes',
 'sourcesquirrel.constants',
 'sourcesquirrel.integrations',
 'sourcesquirrel.integrations.google',
 'sourcesquirrel.prefabs',
 'sourcesquirrel.utils',
 'sourcesquirrel.validators']

package_data = \
{'': ['*']}

install_requires = \
['openpyxl>=3.1.5,<4.0.0', 'pydrive2>=1.21.3,<2.0.0']

setup_kwargs = {
    'name': 'bookio-sourcesquirrel',
    'version': '0.3.0',
    'description': 'Source of truth for book.io and stuff.io data',
    'long_description': '# book.io / sourcesquirrel\n\n---\n\n![source, the squirrel](/img/source.png)\n> source, the squirrel 🐿\n\n',
    'author': 'Fede',
    'author_email': 'fede@book.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/book-io/sourcesquirrel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from sourcesquirrel.helpers import required


class Platform:
    def __init__(self, name: str):
        self.id: str = name
        self.name: str = name

    @property
    def url(self) -> str:
        return f"https://{self.name}"

    def verify(self):
        required("platform.name", self.name, str)

    def __str__(self) -> str:  # pragma: no cover
        return self.name



from datetime import date
from typing import List, Optional
from uuid import UUID

from sourcesquirrel.classes.blockchain import Blockchain
from sourcesquirrel.classes.media_type import MediaType
from sourcesquirrel.classes.release_type import ReleaseType
from sourcesquirrel.classes.serie import Serie
from sourcesquirrel.helpers import required


class Collection:
    def __init__(
        self,
        id: UUID,
        title: str,
        onchain_collection_id: str,
        onchain_collection_id_mainnet: str,
        is_public_domain: bool,
        release_date: date,
        supply: int,
        cover_count: int,
        one_to_one_count: int,
        source_royalties: float,
        team_royalties: float,
        blockchain: Blockchain,
        media_type: MediaType,
        release_type: ReleaseType,
        onchain_discriminator: str = None,
        authors: List[str] = None,
        publishers: List[str] = None,
        series: List[Serie] = None,
    ):
        self.id: UUID = id
        self.title: str = title
        self.onchain_collection_id: str = onchain_collection_id
        self.onchain_collection_id_mainnet: str = onchain_collection_id_mainnet
        self.is_public_domain: bool = is_public_domain
        self.release_date: date = release_date
        self.supply: int = supply
        self.cover_count: int = cover_count
        self.one_to_one_count: int = one_to_one_count
        self.source_royalties: float = source_royalties
        self.team_royalties: float = team_royalties
        self.blockchain: Blockchain = blockchain
        self.media_type: MediaType = media_type
        self.release_type: ReleaseType = release_type
        self.onchain_discriminator: Optional[str] = onchain_discriminator
        self.authors: List[str] = authors or []
        self.publishers: List[str] = publishers or []
        self.series: List[Serie] = series or []

    def verify(self):
        required("collection.id", self.id, UUID)
        required("collection.title", self.title, str)
        required("collection.onchain_collection_id", self.onchain_collection_id, str)
        required("collection.onchain_collection_id_mainnet", self.onchain_collection_id_mainnet, str)
        required("collection.is_public_domain", self.is_public_domain, bool)
        required("collection.release_date", self.release_date, date)
        required("collection.supply", self.supply, int)
        required("collection.cover_count", self.cover_count, int)
        required("collection.one_to_one_count", self.one_to_one_count, int)
        required("collection.source_royalties", self.source_royalties, float)
        required("collection.team_royalties", self.team_royalties, float)
        required("collection.blockchain", self.blockchain, Blockchain)
        required("collection.media_type", self.media_type, MediaType)
        required("collection.release_type", self.release_type, ReleaseType)

        for i, author in enumerate(self.authors):
            required(f"collection.authors[{i}]", author, str)

        for i, publisher in enumerate(self.publishers):
            required(f"collection.publishers[{i}]", publisher, str)

        self.blockchain.verify()
        self.media_type.verify()
        self.release_type.verify()

        for i, serie in enumerate(self.series):
            required(f"collection.series[{i}]", serie, Serie)
            serie.verify()

        assert self.blockchain.is_collection_id(
            self.onchain_collection_id,
        ), f"Invalid onchain_collection_id for blockchain {str(self)}"

        assert self.blockchain.is_collection_id(
            self.onchain_collection_id_mainnet,
        ), f"Invalid onchain_collection_id_mainnet for blockchain {str(self)}"

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.media_type.emoji} {self.title} ({self.blockchain.name})"
